<?php
/**
 * Magento
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/osl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@magento.com so we can send you a copy immediately.
 *
 * DISCLAIMER
 *
 * Do not edit or add to this file if you wish to upgrade Magento to newer
 * versions in the future. If you wish to customize Magento for your
 * needs please refer to http://www.magento.com for more information.
 *
 * @category    Mage
 * @package     Mage_Usa
 * @copyright   Copyright (c) 2006-2020 Magento, Inc. (http://www.magento.com)
 * @license     http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */

/**
 * UPS shipping implementation
 *
 * @category    Mage
 * @package     Mage_Usa
 * @author      Magento Core Team <core@magentocommerce.com>
 */
class Mage_Usa_Model_Shipping_Carrier_Ups
    extends Mage_Usa_Model_Shipping_Carrier_Abstract
    implements Mage_Shipping_Model_Carrier_Interface
{

    /**
     * Code of the carrier
     *
     * @var string
     */
    const CODE = 'ups';

    const URL_SANDBOX = 'https://wwwcie.ups.com/';
    const URL_PRODUCTION = 'https://onlinetools.ups.com/';
    const URL_PATH_AUTH = 'security/v1/oauth/token';

    /**
     * Delivery Confirmation level based on origin/destination
     *
     * @var int
     */
    const DELIVERY_CONFIRMATION_SHIPMENT = 1;
    const DELIVERY_CONFIRMATION_PACKAGE = 2;

    /**
     * Code of the carrier
     *
     * @var string
     */
    protected $_code = self::CODE;

    /**
     * Rate request data
     *
     * @var Mage_Shipping_Model_Rate_Request|null
     */
    protected $_request = null;

    /**
     * Raw rate request data
     *
     * @var Varien_Object|null
     */
    protected $_rawRequest = null;

    /**
     * Rate result data
     *
     * @var Mage_Shipping_Model_Rate_Result|null
     */
    protected $_result = null;

    /**
     * Base currency rate
     *
     * @var double
     */
    protected $_baseCurrencyRate;

    protected $_token = null;

    protected function getToken()
    {
        if ($this->_token) {
            return $this->_token;
        }

        $debugData = ['request' => null, 'result' => null];

        try {
            $username = $this->getConfigData('username');
            $password = $this->getConfigData('password');

            $payload = [
                'grant_type' => 'client_credentials'
            ];

            $headers = [
                'Content-Type: application/x-www-form-urlencoded',
                'Authorization: Basic ' . base64_encode("$username:$password")
            ];

            $curl = $this->getCurl(self::URL_PATH_AUTH, http_build_query($payload), $headers);

            $debugData['request'] = json_encode(['url' => curl_getinfo($curl, CURLINFO_EFFECTIVE_URL), 'payload' => $payload, 'headers' => $headers]);
            $response = curl_exec($curl);
            $debugData['result'] = $response;
            $error = curl_error($curl);
            curl_close($curl);

            if ($error) {
                $debugData['result'] = [
                    'error' => $error
                ];
            } else {
                $response = json_decode($response, true);

                if (isset($response['response']['errors'])) {
                    foreach ($response['response']['errors'] as $error) {
                        throw new Exception($error['message'], $error['code']);
                    }
                } else {
                    $this->_token = $response['access_token'];
                }
            }
        } catch (Exception $e) {
            $debugData['result'] = [
                'error' => $e->getMessage(),
                'code' => $e->getCode()
            ];
        }
        $this->_debug($debugData);

        return $this->_token;
    }

    protected function getCurl($url, $payload, $httpHeaders)
    {
        $url = ($this->getConfigFlag('mode_xml') ? self::URL_SANDBOX : self::URL_PRODUCTION) . $url;

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $payload);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $httpHeaders);
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, $payload ? 'POST' : 'GET');
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, $this->getConfigFlag('verify_peer'));

        return $curl;
    }

    /**
     * Get allowed shipping methods
     *
     * @return array
     */
    public function getAllowedMethods()
    {
        $allowed = explode(',', $this->getConfigData('allowed_methods'));
        $arr = [];

        foreach ($allowed as $k) {
            $arr[$k] = $this->getShipmentByCode($k);
        }

        return $arr;
    }

    /**
     * Get shipment by code
     *
     * @param string $code
     * @param string $origin
     * @return array|bool
     */
    public function getShipmentByCode($code, $origin = null)
    {
        if ($origin === null) {
            $origin = $this->getConfigData('origin_shipment');
        }

        $arr = $this->getCode('originShipment', $origin);

        return isset($arr[$code]) ? $arr[$code] : false;
    }

    /**
     * Get configuration data of carrier
     *
     * @param string $type
     * @param string $code
     * @return array|bool
     */
    public function getCode($type, $code = '')
    {
        $codes = [
            'action' => [
                'single' => '3',
                'all' => '4',
            ],

            'originShipment' => [
                // United States Domestic Shipments
                'United States Domestic Shipments' => [
                    '01' => Mage::helper('usa')->__('UPS Next Day Air'),
                    '02' => Mage::helper('usa')->__('UPS Second Day Air'),
                    '03' => Mage::helper('usa')->__('UPS Ground'),
                    '07' => Mage::helper('usa')->__('UPS Worldwide Express'),
                    '08' => Mage::helper('usa')->__('UPS Worldwide Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '12' => Mage::helper('usa')->__('UPS Three-Day Select'),
                    '13' => Mage::helper('usa')->__('UPS Next Day Air Saver'),
                    '14' => Mage::helper('usa')->__('UPS Next Day Air Early A.M.'),
                    '54' => Mage::helper('usa')->__('UPS Worldwide Express Plus'),
                    '59' => Mage::helper('usa')->__('UPS Second Day Air A.M.'),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                ],
                // Shipments Originating in United States
                'Shipments Originating in United States' => [
                    '01' => Mage::helper('usa')->__('UPS Next Day Air'),
                    '02' => Mage::helper('usa')->__('UPS Second Day Air'),
                    '03' => Mage::helper('usa')->__('UPS Ground'),
                    '07' => Mage::helper('usa')->__('UPS Worldwide Express'),
                    '08' => Mage::helper('usa')->__('UPS Worldwide Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '12' => Mage::helper('usa')->__('UPS Three-Day Select'),
                    '14' => Mage::helper('usa')->__(
                        'UPS Next Day Air Early A.M.'
                    ),
                    '54' => Mage::helper('usa')->__(
                        'UPS Worldwide Express Plus'
                    ),
                    '59' => Mage::helper('usa')->__('UPS Second Day Air A.M.'),
                    '65' => Mage::helper('usa')->__('UPS Worldwide Saver'),
                ],
                // Shipments Originating in Canada
                'Shipments Originating in Canada' => [
                    '01' => Mage::helper('usa')->__('UPS Express'),
                    '02' => Mage::helper('usa')->__('UPS Expedited'),
                    '07' => Mage::helper('usa')->__('UPS Worldwide Express'),
                    '08' => Mage::helper('usa')->__('UPS Worldwide Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '12' => Mage::helper('usa')->__('UPS Three-Day Select'),
                    '14' => Mage::helper('usa')->__('UPS Express Early A.M.'),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                ],
                // Shipments Originating in the European Union
                'Shipments Originating in the European Union' => [
                    '07' => Mage::helper('usa')->__('UPS Express'),
                    '08' => Mage::helper('usa')->__('UPS Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '54' => Mage::helper('usa')->__(
                        'UPS Worldwide Express PlusSM'
                    ),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                ],
                // Polish Domestic Shipments
                'Polish Domestic Shipments' => [
                    '07' => Mage::helper('usa')->__('UPS Express'),
                    '08' => Mage::helper('usa')->__('UPS Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '54' => Mage::helper('usa')->__(
                        'UPS Worldwide Express Plus'
                    ),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                    '82' => Mage::helper('usa')->__('UPS Today Standard'),
                    '83' => Mage::helper('usa')->__(
                        'UPS Today Dedicated Courrier'
                    ),
                    '84' => Mage::helper('usa')->__('UPS Today Intercity'),
                    '85' => Mage::helper('usa')->__('UPS Today Express'),
                    '86' => Mage::helper('usa')->__('UPS Today Express Saver'),
                ],
                // Puerto Rico Origin
                'Puerto Rico Origin' => [
                    '01' => Mage::helper('usa')->__('UPS Next Day Air'),
                    '02' => Mage::helper('usa')->__('UPS Second Day Air'),
                    '03' => Mage::helper('usa')->__('UPS Ground'),
                    '07' => Mage::helper('usa')->__('UPS Worldwide Express'),
                    '08' => Mage::helper('usa')->__('UPS Worldwide Expedited'),
                    '14' => Mage::helper('usa')->__(
                        'UPS Next Day Air Early A.M.'
                    ),
                    '54' => Mage::helper('usa')->__(
                        'UPS Worldwide Express Plus'
                    ),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                ],
                // Shipments Originating in Mexico
                'Shipments Originating in Mexico' => [
                    '07' => Mage::helper('usa')->__('UPS Express'),
                    '08' => Mage::helper('usa')->__('UPS Expedited'),
                    '54' => Mage::helper('usa')->__('UPS Express Plus'),
                    '65' => Mage::helper('usa')->__('UPS Saver'),
                ],
                // Shipments Originating in Other Countries
                'Shipments Originating in Other Countries' => [
                    '07' => Mage::helper('usa')->__('UPS Express'),
                    '08' => Mage::helper('usa')->__('UPS Worldwide Expedited'),
                    '11' => Mage::helper('usa')->__('UPS Standard'),
                    '54' => Mage::helper('usa')->__(
                        'UPS Worldwide Express Plus'
                    ),
                    '65' => Mage::helper('usa')->__('UPS Saver')
                ]
            ],

            'method' => [
                '1DM' => Mage::helper('usa')->__('Next Day Air Early AM'),
                '1DML' => Mage::helper('usa')->__('Next Day Air Early AM Letter'),
                '1DA' => Mage::helper('usa')->__('Next Day Air'),
                '1DAL' => Mage::helper('usa')->__('Next Day Air Letter'),
                '1DAPI' => Mage::helper('usa')->__('Next Day Air Intra (Puerto Rico)'),
                '1DP' => Mage::helper('usa')->__('Next Day Air Saver'),
                '1DPL' => Mage::helper('usa')->__('Next Day Air Saver Letter'),
                '2DM' => Mage::helper('usa')->__('2nd Day Air AM'),
                '2DML' => Mage::helper('usa')->__('2nd Day Air AM Letter'),
                '2DA' => Mage::helper('usa')->__('2nd Day Air'),
                '2DAL' => Mage::helper('usa')->__('2nd Day Air Letter'),
                '3DS' => Mage::helper('usa')->__('3 Day Select'),
                'GND' => Mage::helper('usa')->__('Ground'),
                'GNDCOM' => Mage::helper('usa')->__('Ground Commercial'),
                'GNDRES' => Mage::helper('usa')->__('Ground Residential'),
                'STD' => Mage::helper('usa')->__('Canada Standard'),
                'XPR' => Mage::helper('usa')->__('Worldwide Express'),
                'WXS' => Mage::helper('usa')->__('Worldwide Express Saver'),
                'XPRL' => Mage::helper('usa')->__('Worldwide Express Letter'),
                'XDM' => Mage::helper('usa')->__('Worldwide Express Plus'),
                'XDML' => Mage::helper('usa')->__('Worldwide Express Plus Letter'),
                'XPD' => Mage::helper('usa')->__('Worldwide Expedited'),
            ],

            'pickup' => [
                'RDP' => ["label" => 'Regular Daily Pickup', "code" => "01"],
                'OCA' => ["label" => 'On Call Air', "code" => "07"],
                'OTP' => ["label" => 'One Time Pickup', "code" => "06"],
                'LC' => ["label" => 'Letter Center', "code" => "19"],
                'CC' => ["label" => 'Customer Counter', "code" => "03"],
            ],

            'container' => [
                'CP' => '00', // Customer Packaging
                'ULE' => '01', // UPS Letter Envelope
                'CSP' => '02', // Customer Supplied Package
                'UT' => '03', // UPS Tube
                'PAK' => '04', // PAK
                'UEB' => '21', // UPS Express Box
                'UW25' => '24', // UPS Worldwide 25 kilo
                'UW10' => '25', // UPS Worldwide 10 kilo
                'PLT' => '30', // Pallet
                'SEB' => '2a', // Small Express Box
                'MEB' => '2b', // Medium Express Box
                'LEB' => '2c', // Large Express Box
            ],

            'container_description' => [
                'CP' => Mage::helper('usa')->__('Customer Packaging'),
                'ULE' => Mage::helper('usa')->__('UPS Letter Envelope'),
                'CSP' => Mage::helper('usa')->__('Customer Supplied Package'),
                'UT' => Mage::helper('usa')->__('UPS Tube'),
                'PAK' => Mage::helper('usa')->__('PAK'),
                'UEB' => Mage::helper('usa')->__('UPS Express Box'),
                'UW25' => Mage::helper('usa')->__('UPS Worldwide 25 kilo'),
                'UW10' => Mage::helper('usa')->__('UPS Worldwide 10 kilo'),
                'PLT' => Mage::helper('usa')->__('Pallet'),
                'SEB' => Mage::helper('usa')->__('Small Express Box'),
                'MEB' => Mage::helper('usa')->__('Medium Express Box'),
                'LEB' => Mage::helper('usa')->__('Large Express Box'),
            ],

            'dest_type' => [
                'RES' => '01', // Residential
                'COM' => '02', // Commercial
            ],

            'dest_type_description' => [
                'RES' => Mage::helper('usa')->__('Residential'),
                'COM' => Mage::helper('usa')->__('Commercial'),
            ],

            'unit_of_measure' => [
                'LBS' => Mage::helper('usa')->__('Pounds'),
                'KGS' => Mage::helper('usa')->__('Kilograms'),
            ],

            'containers_filter' => [
                [
                    'containers' => ['00'], // Customer Packaging
                    'filters' => [
                        'within_us' => [
                            'method' => [
                                '01', // Next Day Air
                                '13', // Next Day Air Saver
                                '12', // 3 Day Select
                                '59', // 2nd Day Air AM
                                '03', // Ground
                                '14', // Next Day Air Early AM
                                '02', // 2nd Day Air
                            ]
                        ],
                        'from_us' => [
                            'method' => [
                                '07', // Worldwide Express
                                '54', // Worldwide Express Plus
                                '08', // Worldwide Expedited
                                '65', // Worldwide Saver
                                '11', // Standard
                            ]
                        ]
                    ]
                ],
                [
                    // Small Express Box, Medium Express Box, Large Express Box, UPS Tube
                    'containers' => ['2a', '2b', '2c', '03'],
                    'filters' => [
                        'within_us' => [
                            'method' => [
                                '01', // Next Day Air
                                '13', // Next Day Air Saver
                                '14', // Next Day Air Early AM
                                '02', // 2nd Day Air
                                '59', // 2nd Day Air AM
                                '13', // Next Day Air Saver
                            ]
                        ],
                        'from_us' => [
                            'method' => [
                                '07', // Worldwide Express
                                '54', // Worldwide Express Plus
                                '08', // Worldwide Expedited
                                '65', // Worldwide Saver
                            ]
                        ]
                    ]
                ],
                [
                    'containers' => ['24', '25'],
                    // UPS Worldwide 25 kilo, UPS Worldwide 10 kilo
                    'filters' => [
                        'within_us' => [
                            'method' => []
                        ],
                        'from_us' => [
                            'method' => [
                                '07', // Worldwide Express
                                '54', // Worldwide Express Plus
                                '65', // Worldwide Saver
                            ]
                        ]
                    ]
                ],
                [
                    'containers' => ['01', '04'], // UPS Letter, UPS PAK
                    'filters' => [
                        'within_us' => [
                            'method' => [
                                '01', // Next Day Air
                                '14', // Next Day Air Early AM
                                '02', // 2nd Day Air
                                '59', // 2nd Day Air AM
                                '13', // Next Day Air Saver
                            ]
                        ],
                        'from_us' => [
                            'method' => [
                                '07', // Worldwide Express
                                '54', // Worldwide Express Plus
                                '65', // Worldwide Saver
                            ]
                        ]
                    ]
                ],
                [
                    'containers' => ['04'], // UPS PAK
                    'filters' => [
                        'within_us' => [
                            'method' => []
                        ],
                        'from_us' => [
                            'method' => [
                                '08', // Worldwide Expedited
                            ]
                        ]
                    ]
                ],
            ]
        ];

        if (!isset($codes[$type])) {
            return false;
        } elseif ('' === $code) {
            return $codes[$type];
        }

        if (!isset($codes[$type][$code])) {
            return false;
        } else {
            return $codes[$type][$code];
        }
    }

    /**
     * Collect and get rates
     *
     * @param Mage_Shipping_Model_Rate_Request $request
     * @return Mage_Shipping_Model_Rate_Result|bool|null
     */
    public function collectRates(Mage_Shipping_Model_Rate_Request $request)
    {
        if (!$this->getConfigFlag($this->_activeFlag)) {
            return false;
        }

        $this->setRequest($request);

        $this->_result = $this->getQuotes();
        $this->_updateFreeMethodQuote($request);

        return $this->_result;
    }

    /**
     * Prepare and set request to this instance
     *
     * @param Mage_Shipping_Model_Rate_Request $request
     * @return Mage_Usa_Model_Shipping_Carrier_Ups
     */
    public function setRequest(Mage_Shipping_Model_Rate_Request $request)
    {
        $this->_request = $request;

        $r = new Varien_Object();

        if ($request->getLimitMethod()) {
            $r->setAction($this->getCode('action', 'single'));
            $r->setProduct($request->getLimitMethod());
        } else {
            $r->setAction($this->getCode('action', 'all'));
            $r->setProduct('GND' . $this->getConfigData('dest_type'));
        }

        if ($request->getUpsPickup()) {
            $pickup = $request->getUpsPickup();
        } else {
            $pickup = $this->getConfigData('pickup');
        }
        $r->setPickup($this->getCode('pickup', $pickup));

        if ($request->getUpsContainer()) {
            $container = $request->getUpsContainer();
        } else {
            $container = $this->getConfigData('container');
        }
        $r->setContainer($this->getCode('container', $container));

        if ($request->getUpsDestType()) {
            $destType = $request->getUpsDestType();
        } else {
            $destType = $this->getConfigData('dest_type');
        }
        $r->setDestType($this->getCode('dest_type', $destType));

        if ($request->getOrigCountry()) {
            $origCountry = $request->getOrigCountry();
        } else {
            $origCountry = Mage::getStoreConfig(
                Mage_Shipping_Model_Shipping::XML_PATH_STORE_COUNTRY_ID,
                $request->getStoreId()
            );
        }

        $r->setOrigCountry(Mage::getModel('directory/country')->load($origCountry)->getIso2Code());

        if ($request->getOrigRegionCode()) {
            $origRegionCode = $request->getOrigRegionCode();
        } else {
            $origRegionCode = Mage::getStoreConfig(
                Mage_Shipping_Model_Shipping::XML_PATH_STORE_REGION_ID,
                $request->getStoreId()
            );
        }
        if (is_numeric($origRegionCode)) {
            $origRegionCode = Mage::getModel('directory/region')->load($origRegionCode)->getCode();
        }
        $r->setOrigRegionCode($origRegionCode);

        if ($request->getOrigPostcode()) {
            $r->setOrigPostal($request->getOrigPostcode());
        } else {
            $r->setOrigPostal(Mage::getStoreConfig(
                Mage_Shipping_Model_Shipping::XML_PATH_STORE_ZIP,
                $request->getStoreId()
            ));
        }

        if ($request->getOrigCity()) {
            $r->setOrigCity($request->getOrigCity());
        } else {
            $r->setOrigCity(Mage::getStoreConfig(
                Mage_Shipping_Model_Shipping::XML_PATH_STORE_CITY,
                $request->getStoreId()
            ));
        }

        if ($request->getDestCountryId()) {
            $destCountry = $request->getDestCountryId();
        } else {
            $destCountry = self::USA_COUNTRY_ID;
        }

        //for UPS, puerto rico state for US will assume as puerto rico country
        if ($destCountry == self::USA_COUNTRY_ID
            && ($request->getDestPostcode() == '00912' || $request->getDestRegionCode() == self::PUERTORICO_COUNTRY_ID)
        ) {
            $destCountry = self::PUERTORICO_COUNTRY_ID;
        }

        // For UPS, Guam state of the USA will be represented by Guam country
        if ($destCountry == self::USA_COUNTRY_ID && $request->getDestRegionCode() == self::GUAM_REGION_CODE) {
            $destCountry = self::GUAM_COUNTRY_ID;
        }

        $r->setDestCountry(Mage::getModel('directory/country')->load($destCountry)->getIso2Code());

        $r->setDestRegionCode($request->getDestRegionCode());

        if ($request->getDestPostcode()) {
            $r->setDestPostal($request->getDestPostcode());
        }

        $weight = $this->getTotalNumOfBoxes($request->getPackageWeight());

        $weight = $this->_getCorrectWeight($weight);

        $r->setWeight($weight);
        if ($request->getFreeMethodWeight() != $request->getPackageWeight()) {
            $r->setFreeMethodWeight($request->getFreeMethodWeight());
        }

        $r->setValue($request->getPackageValue());
        $r->setValueWithDiscount($request->getPackageValueWithDiscount());

        if ($request->getUpsUnitMeasure()) {
            $unit = $request->getUpsUnitMeasure();
        } else {
            $unit = $this->getConfigData('unit_of_measure');
        }
        $r->setUnitMeasure($unit);

        $r->setIsReturn($request->getIsReturn());

        $r->setBaseSubtotalInclTax($request->getBaseSubtotalInclTax());

        $this->_rawRequest = $r;

        return $this;
    }

    /**
     * Do remote request for  and handle errors
     *
     * @return Mage_Shipping_Model_Rate_Result
     */
    protected function getQuotes()
    {
        /** @var Mage_Shipping_Model_Rate_Result $result */
        $result = Mage::getModel('shipping/rate_result');

        $debugData = ['request' => null, 'result' => null];

        try {
            $r = $this->_rawRequest;

            if ($r->getAction() == '4') {
                $requestOption = 'Shop';
                $serviceCode = null; // Service code is not relevant when we're asking ALL possible services' rates
            } else {
                $requestOption = 'Rate';
                $serviceCode = $r->getProduct() ? $r->getProduct() : '';
            }
            $serviceDescription = $serviceCode ? $this->getShipmentByCode($serviceCode) : '';

            $srcPostal = $r->getOrigPostal();
            $srcCountryCode = $r->getOrigCountry();
            $srcStateProvince = $r->getOrigRegionCode();

            $destPostal = Mage_Usa_Model_Shipping_Carrier_Abstract::USA_COUNTRY_ID == $r->getDestCountry() ? substr($r->getDestPostal(), 0, 5) : $r->getDestPostal();
            $destCountry = $r->getDestCountry();
            $destRegionCode = $r->getDestRegionCode();

            if ($r->getIsReturn()) {
                $shipperCity = '';
                $shipperPostalCode = $destPostal;
                $shipperCountryCode = $destCountry;
                $shipperStateProvince = $destRegionCode;
            } else {
                $shipperCity = $r->getOrigCity();
                $shipperPostalCode = $srcPostal;
                $shipperCountryCode = $srcCountryCode;
                $shipperStateProvince = $srcStateProvince;
            }

            $negotiatedActive = $this->getConfigFlag('negotiated_active');

            $payload = [
                'RateRequest' => [
                    'Request' => [
                        'TransactionReference' => [
                            'CustomerContext' => 'Rating and Service'
                        ],
                        'RequestOption' => $requestOption,
                    ],
                    'Shipment' => [
                        'Shipper' => [
                            'ShipperNumber' => ($negotiatedActive && ($shipper = $this->getConfigData('shipper_number'))) ? $shipper : '',
                            'Address' => [
                                'City' => $shipperCity,
                                'StateProvinceCode' => $shipperStateProvince,
                                'PostalCode' => $shipperPostalCode,
                                'CountryCode' => $shipperCountryCode
                            ],
                        ],
                        'ShipTo' => [
                            'Address' => [
                                'StateProvinceCode' => $destRegionCode,
                                'PostalCode' => $destPostal,
                                'CountryCode' => $destCountry
                            ]
                        ],
                        'ShipFrom' => [
                            'Address' => [
                                'StateProvinceCode' => $srcStateProvince,
                                'PostalCode' => $srcPostal,
                                'CountryCode' => $srcCountryCode
                            ]
                        ],
                        'Package' => [
                            'PackagingType' => [
                                'Code' => $r->getContainer(),
                            ],
                            'PackageWeight' => [
                                'UnitOfMeasurement' => [
                                    'Code' => $r->getUnitMeasure(),
                                ],
                                'Weight' => strval($r->getWeight())
                            ]
                        ],
                        'ShipmentRatingOptions' => [
                            'NegotiatedRatesIndicator' => $negotiatedActive ? 'true' : 'false'
                        ]
                    ]
                ]
            ];

            if ($serviceCode !== null) {
                $payload['RateRequest']['Shipment']['Service']['Code'] = $serviceCode;
                $payload['RateRequest']['Shipment']['Service']['Description'] = $serviceDescription;
            }

            $headers = [
                'Content-Type: application/json',
                'Authorization: Bearer ' . $this->getToken()
            ];

            $curl = $this->getCurl('api/rating/v2403/' . $requestOption, json_encode($payload), $headers);

            $debugData['request'] = json_encode(['url' => curl_getinfo($curl, CURLINFO_EFFECTIVE_URL), 'payload' => $payload, 'headers' => $headers]);
            $response = curl_exec($curl);
            $debugData['result'] = $response;
            $error = curl_error($curl);
            curl_close($curl);

            if ($error) {
                $debugData['result'] = [
                    'error' => $error
                ];
            } else {
                $response = json_decode($response, true);

                if (isset($response['response']['errors'])) {
                    foreach ($response['response']['errors'] as $error) {
                        $debugData['result'] = [
                            'error' => $error['message'],
                            'code' => $error['code']
                        ];

                        $error = Mage::getModel('shipping/rate_result_error');
                        $error->setCarrier(self::CODE);
                        $error->setCarrierTitle($this->getConfigData('title'));
                        $error->setErrorMessage($error['message']);
                        $result->append($error);
                    }
                } else {
                    $costArr = [];
                    $priceArr = [];
                    $arr = $response['RateResponse']['RatedShipment'];
                    $allowedMethods = explode(',', $this->getConfigData('allowed_methods'));
                    $allowedCurrencies = Mage::getModel('directory/currency')->getConfigAllowCurrencies();

                    foreach ($arr as $shipElement) {
                        $code = (string)$shipElement['Service']['Code'];

                        if (in_array($code, $allowedMethods)) {
                            $negotiatedActive = $this->getConfigFlag('negotiated_active')
                                && $this->getConfigData('shipper_number')
                                && isset($shipElement['NegotiatedRateCharges']);

                            if ($negotiatedActive) {
                                $cost = $shipElement['NegotiatedRates']['TotalCharge']['MonetaryValue'];
                            } else {
                                $cost = $shipElement['TotalCharges']['MonetaryValue'];
                            }

                            //convert price with Origin country currency code to base currency code
                            $successConversion = true;
                            $responseCurrencyCode = (string)$shipElement['TotalCharges']['CurrencyCode'];

                            if ($responseCurrencyCode) {
                                if (in_array($responseCurrencyCode, $allowedCurrencies)) {
                                    $cost = (float)$cost * $this->_getBaseCurrencyRate($responseCurrencyCode);
                                } else {
                                    $errorTitle = Mage::helper('directory')->__('Can\'t convert rate from "%s-%s".', $responseCurrencyCode, $this->_request->getPackageCurrency()->getCode());
                                    $error = Mage::getModel('shipping/rate_result_error');
                                    $error->setCarrier(self::CODE);
                                    $error->setCarrierTitle($this->getConfigData('title'));
                                    $error->setErrorMessage($errorTitle);
                                    $successConversion = false;
                                }
                            }

                            if ($successConversion) {
                                $costArr[$code] = $cost;
                                $priceArr[$code] = $this->getMethodPrice(floatval($cost), $code);
                            }
                        }
                    }

                    if (empty($priceArr)) {
                        $error = Mage::getModel('shipping/rate_result_error');
                        $error->setCarrier(self::CODE);
                        $error->setCarrierTitle($this->getConfigData('title'));
                        $error->setErrorMessage($this->getConfigData('specificerrmsg'));
                        $result->append($error);
                    } else {
                        foreach ($priceArr as $method => $price) {
                            $method_arr = $this->getShipmentByCode($method);

                            if ($method_arr) {
                                $rate = Mage::getModel('shipping/rate_result_method');
                                $rate->setCarrier(self::CODE);
                                $rate->setCarrierTitle($this->getConfigData('title'));
                                $rate->setMethod($method);
                                $rate->setMethodTitle($method_arr);
                                $rate->setCost($costArr[$method]);
                                $rate->setPrice($price);
                                $result->append($rate);
                            }
                        }
                    }
                }
            }
        } catch (Exception $e) {
            $debugData['result'] = [
                'error' => $e->getMessage(),
                'code' => $e->getCode()
            ];

            $error = Mage::getModel('shipping/rate_result_error');
            $error->setCarrier(self::CODE);
            $error->setCarrierTitle($this->getConfigData('title'));
            $error->setErrorMessage($e->getMessage());
            $result->append($error);
        }
        $this->_debug($debugData);

        return $result;
    }

    /**
     * Get correct weigt.
     *
     * Namely:
     * Checks the current weight to comply with the minimum weight standards set by the carrier.
     * Then strictly rounds the weight up until the first significant digit after the decimal point.
     *
     * @param float|integer|double $weight
     * @return float
     */
    protected function _getCorrectWeight($weight)
    {
        $minWeight = $this->getConfigData('min_package_weight');

        if ($weight < $minWeight) {
            $weight = $minWeight;
        }

        //rounds a number to one significant figure
        $weight = ceil($weight * 10) / 10;

        return $weight;
    }

    /**
     * Get base currency rate
     *
     * @param string $code
     * @return double
     */
    protected function _getBaseCurrencyRate($code)
    {
        if (!$this->_baseCurrencyRate) {
            $this->_baseCurrencyRate = Mage::getModel('directory/currency')
                ->load($code)
                ->getAnyRate($this->_request->getBaseCurrency()->getCode());
        }

        return $this->_baseCurrencyRate;
    }

    /**
     * Do shipment request to carrier web service, obtain Print Shipping Labels and process errors in response
     *
     * @param Varien_Object $request
     * @return Varien_Object
     */
    public function _doShipmentRequest(Varien_Object $request)
    {
        $this->_prepareShipmentRequest($request);

        $packageParams = $request->getPackageParams();
        $height = $packageParams->getHeight();
        $width = $packageParams->getWidth();
        $length = $packageParams->getLength();
        $weightUnits = $packageParams->getWeightUnits() == Zend_Measure_Weight::POUND ? 'LBS' : 'KGS';
        $dimensionsUnits = $packageParams->getDimensionUnits() == Zend_Measure_Length::INCH ? 'IN' : 'CM';

        $itemsDesc = array();
        $itemsShipment = $request->getPackageItems();
        foreach ($itemsShipment as $itemShipment) {
            $item = new Varien_Object();
            $item->setData($itemShipment);
            $itemsDesc[] = $item->getName();
        }

        $payload = [
            'ShipmentRequest' => [
                'Shipment' => [
                    'Description' => substr(implode(' ', $itemsDesc), 0, 35),
                    'ShipTo' => [
                        'Name' => $request->getRecipientContactCompanyName() ? $request->getRecipientContactCompanyName() : 'N/A',
                        'AttentionName' => $request->getRecipientContactPersonName(),
                        'Phone' => [
                            'Number' => $request->getRecipientContactPhoneNumber()
                        ],
                        'Address' => [
                            'AddressLine' => [
                                $request->getRecipientAddressStreet1(),
                                $request->getRecipientAddressStreet2(),
                            ],
                            'City' => $request->getRecipientAddressCity(),
                            'PostalCode' => $request->getRecipientAddressPostalCode(),
                            'CountryCode' => $request->getRecipientAddressCountryCode(),
                            'StateProvinceCode' => ($request->getRecipientAddressStateOrProvinceCode()) ? $request->getRecipientAddressRegionCode() : ''
                        ],
                    ],
                    'PaymentInformation' => [
                        'ShipmentCharge' => [
                            [
                                'Type' => '01',
                                'BillShipper' => [
                                    'AccountNumber' => $this->getConfigData('shipper_number')
                                ]
                            ]
                        ]
                    ],
                    'Service' => [
                        'Code' => $request->getShippingMethod()
                    ],
                    'Package' => [
                        'Description' => substr(implode(' ', $itemsDesc), 0, 35),
                        'Packaging' => [
                            'Code' => $request->getPackagingType()
                        ],
                        'PackageWeight' => [
                            'UnitOfMeasurement' => [
                                'Code' => $weightUnits
                            ],
                            'Weight' => $request->getPackageWeight()
                        ],
                    ]
                ],
                'LabelSpecification' => [
                    'LabelImageFormat' => [
                        'Code' => 'GIF',
                        'Description' => 'GIF'
                    ]
                ]
            ]
        ];

        if ($request->getIsReturn()) {
            $payload['ShipmentRequest']['Shipment']['ReturnService'] = [
                'Code' => '9',
                'Description' => 'UPS Print Return Label (PRL)'
            ];
            $payload['ShipmentRequest']['Shipment']['Shipper']['Name'] = $request->getRecipientContactCompanyName();
            $payload['ShipmentRequest']['Shipment']['Shipper']['AttentionName'] = $request->getRecipientContactPersonName();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Phone']['Number'] = $request->getRecipientContactPhoneNumber();
            $payload['ShipmentRequest']['Shipment']['Shipper']['ShipperNumber'] = $this->getConfigData('shipper_number');

            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['AddressLine'] = [
                $request->getRecipientAddressStreet(),
                $request->getRecipientAddressStreet2()
            ];
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['City'] = $request->getRecipientAddressCity();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['CountryCode'] = $request->getRecipientAddressCountryCode();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['PostalCode'] = $request->getRecipientAddressPostalCode();
            if ($request->getRecipientAddressStateOrProvinceCode()) {
                $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['StateProvinceCode'] = $request->getRecipientAddressStateOrProvinceCode();
            }
        } else {
            $payload['ShipmentRequest']['Shipment']['Shipper']['Name'] = $request->getShipperContactCompanyName();
            $payload['ShipmentRequest']['Shipment']['Shipper']['AttentionName'] = $request->getShipperContactPersonName();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Phone']['Number'] = $request->getShipperContactPhoneNumber();
            $payload['ShipmentRequest']['Shipment']['Shipper']['ShipperNumber'] = $this->getConfigData('shipper_number');

            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['AddressLine'] = [
                $request->getShipperAddressStreet(),
                $request->getShipperAddressStreet2()
            ];
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['City'] = $request->getShipperAddressCity();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['CountryCode'] = $request->getShipperAddressCountryCode();
            $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['PostalCode'] = $request->getShipperAddressPostalCode();
            if ($request->getShipperAddressStateOrProvinceCode()) {
                $payload['ShipmentRequest']['Shipment']['Shipper']['Address']['StateProvinceCode'] = $request->getShipperAddressStateOrProvinceCode();
            }
        }

        if ($this->getConfigData('dest_type') == 'RES') {
            $payload['ShipmentRequest']['Shipment']['ShipTo']['Address']['ResidentialAddressIndicator'] = ' ';
        }

        if ($request->getIsReturn()) {
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Name'] = $request->getShipperContactCompanyName() ? $request->getShipperContactCompanyName() : $request->getShipperContactPersonName();
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['AttentionName'] = $request->getShipperContactPersonName();
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Phone']['Number'] = $request->getShipperContactPhoneNumber();
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['AddressLine'] = [
                $request->getShipperAddressStreet(),
                $request->getShipperAddressStreet2()
            ];
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['City'] = $request->getShipperAddressCity();
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['CountryCode'] = $request->getShipperAddressCountryCode();
            $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['PostalCode'] = $request->getShipperAddressPostalCode();
            if ($request->getShipperAddressStateOrProvinceCode()) {
                $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['StateProvinceCode'] = $request->getShipperAddressStateOrProvinceCode();
            }
            if ($this->getConfigData('dest_type') == 'RES') {
                $payload['ShipmentRequest']['Shipment']['ShipFrom']['Address']['ResidentialAddressIndicator'] = ' ';
            }
        }

        if ($length || $width || $height) {
            $payload['ShipmentRequest']['Shipment']['Package']['Dimensions']['UnitOfMeasurement']['Code'] = $dimensionsUnits;
            $payload['ShipmentRequest']['Shipment']['Package']['Dimensions']['Length'] = $length;
            $payload['ShipmentRequest']['Shipment']['Package']['Dimensions']['Width'] = $width;
            $payload['ShipmentRequest']['Shipment']['Package']['Dimensions']['Height'] = $height;
        }

        if ($this->_isUSCountry($request->getRecipientAddressCountryCode()) && $this->_isUSCountry($request->getShipperAddressCountryCode())) {
            if ($request->getReferenceData()) {
                $referenceData = $request->getReferenceData() . $request->getPackageId();
            } else {
                $referenceData = 'Order #'
                    . $request->getOrderShipment()->getOrder()->getIncrementId()
                    . ' P'
                    . $request->getPackageId();
            }

            $payload['ShipmentRequest']['Shipment']['ReferenceNumber'] = [
                'Code' => '02',
                'Value' => $referenceData
            ];
        }

        if ($deliveryConfirmation = $packageParams->getDeliveryConfirmation()) {
            switch ($this->_getDeliveryConfirmationLevel($request->getRecipientAddressCountryCode())) {
                case self::DELIVERY_CONFIRMATION_PACKAGE:
                    $payload['ShipmentRequest']['Shipment']['Package']['PackageServiceOptions']['DeliveryConfirmation']['DCISType'] = $deliveryConfirmation;
                    break;
                case self::DELIVERY_CONFIRMATION_SHIPMENT:
                    $payload['ShipmentRequest']['Shipment']['ShipmentServiceOptions']['DeliveryConfirmation']['DCISType'] = $deliveryConfirmation;
                    break;
            }
        }

        if ($request->getPackagingType() != $this->getCode('container', 'ULE')
            && $request->getShipperAddressCountryCode() == Mage_Usa_Model_Shipping_Carrier_Abstract::USA_COUNTRY_ID
            && ($request->getRecipientAddressCountryCode() == 'CA' //Canada
                || $request->getRecipientAddressCountryCode() == 'PR' //Puerto Rico
            )) {
            $payload['ShipmentRequest']['Shipment']['InvoiceLineTotal'] = [
                'CurrencyCode' => $request->getBaseCurrencyCode(),
                'MonetaryValue' => ceil($packageParams->getCustomsValue())
            ];
        }

        $debugData = ['request' => null, 'result' => null];
        $result = new Varien_Object();

        try {
            $headers = [
                'Content-Type: application/json',
                'Authorization: Bearer ' . $this->getToken()
            ];

            $curl = $this->getCurl('api/shipments/v2403/ship', json_encode($payload), $headers);

            $debugData['request'] = json_encode(['url' => curl_getinfo($curl, CURLINFO_EFFECTIVE_URL), 'payload' => $payload, 'headers' => $headers]);
            $response = curl_exec($curl);
            $debugData['result'] = $response;
            $error = curl_error($curl);
            curl_close($curl);

            if ($error) {
                $debugData['result'] = [
                    'error' => $error
                ];
            } else {
                $response = json_decode($response, true);

                if (isset($response['response']['errors'])) {
                    foreach ($response['response']['errors'] as $error) {
                        throw new Exception($error['message'], $error['code']);
                    }
                } else {
                    $shippingLabelContent = (string)$response['ShipmentResponse']['ShipmentResults']['PackageResults'][0]['ShippingLabel']['GraphicImage'];
                    $trackingNumber = (string)$response['ShipmentResponse']['ShipmentResults']['PackageResults'][0]['TrackingNumber'];

                    $result->setShippingLabelContent(base64_decode($shippingLabelContent));
                    $result->setTrackingNumber($trackingNumber);
                }
            }
        } catch (Exception $e) {
            $result->setErrors($e->getMessage());
        }
        $this->_debug($debugData);

        return $result;
    }

    /**
     * Return container types of carrier
     *
     * @param Varien_Object|null $params
     * @return array|bool
     */
    public function getContainerTypes(Varien_Object $params = null)
    {
        if ($params == null) {
            return $this->_getAllowedContainers($params);
        }
        $method = $params->getMethod();
        $countryShipper = $params->getCountryShipper();
        $countryRecipient = $params->getCountryRecipient();

        if (($countryShipper == self::USA_COUNTRY_ID
                && $countryRecipient == self::CANADA_COUNTRY_ID)
            || ($countryShipper == self::CANADA_COUNTRY_ID
                && $countryRecipient == self::USA_COUNTRY_ID)
            || ($countryShipper == self::MEXICO_COUNTRY_ID
                && $countryRecipient == self::USA_COUNTRY_ID)
            && $method == '11' // UPS Standard
        ) {
            $containerTypes = array();
            if ($method == '07' // Worldwide Express
                || $method == '08' // Worldwide Expedited
                || $method == '65' // Worldwide Saver

            ) {
                // Worldwide Expedited
                if ($method != '08') {
                    $containerTypes = array(
                        '01' => Mage::helper('usa')->__('UPS Letter Envelope'),
                        '24' => Mage::helper('usa')->__('UPS Worldwide 25 kilo'),
                        '25' => Mage::helper('usa')->__('UPS Worldwide 10 kilo'),
                    );
                }
                $containerTypes = $containerTypes + array(
                        '03' => Mage::helper('usa')->__('UPS Tube'),
                        '04' => Mage::helper('usa')->__('PAK'),
                        '2a' => Mage::helper('usa')->__('Small Express Box'),
                        '2b' => Mage::helper('usa')->__('Medium Express Box'),
                        '2c' => Mage::helper('usa')->__('Large Express Box'),
                    );
            }
            return array('00' => Mage::helper('usa')->__('Customer Packaging')) + $containerTypes;
        } elseif ($countryShipper == self::USA_COUNTRY_ID && $countryRecipient == self::PUERTORICO_COUNTRY_ID
            && ($method == '03' // UPS Ground
                || $method == '02' // UPS Second Day Air
                || $method == '01' // UPS Next Day Air
            )) {
            // Container types should be the same as for domestic
            $params->setCountryRecipient(self::USA_COUNTRY_ID);
            $containerTypes = $this->_getAllowedContainers($params);
            $params->setCountryRecipient($countryRecipient);
            return $containerTypes;
        }
        return $this->_getAllowedContainers($params);
    }

    /**
     * Return all container types of carrier
     *
     * @return array|bool
     */
    public function getContainerTypesAll()
    {
        $codes = $this->getCode('container');
        $descriptions = $this->getCode('container_description');
        $result = array();
        foreach ($codes as $key => $code) {
            $result[$code] = $descriptions[$key];
        }

        return $result;
    }

    /**
     * Return delivery confirmation types of carrier
     *
     * @param Varien_Object|null $params
     * @return array|bool
     */
    public function getDeliveryConfirmationTypes(Varien_Object $params = null)
    {
        $countryRecipient = $params != null ? $params->getCountryRecipient() : null;
        $deliveryConfirmationTypes = array();
        switch ($this->_getDeliveryConfirmationLevel($countryRecipient)) {
            case self::DELIVERY_CONFIRMATION_PACKAGE:
                $deliveryConfirmationTypes = array(
                    1 => Mage::helper('usa')->__('Delivery Confirmation'),
                    2 => Mage::helper('usa')->__('Signature Required'),
                    3 => Mage::helper('usa')->__('Adult Signature Required'),
                );
                break;
            case self::DELIVERY_CONFIRMATION_SHIPMENT:
                $deliveryConfirmationTypes = array(
                    1 => Mage::helper('usa')->__('Signature Required'),
                    2 => Mage::helper('usa')->__('Adult Signature Required'),
                );
        }
        array_unshift($deliveryConfirmationTypes, Mage::helper('usa')->__('Not Required'));

        return $deliveryConfirmationTypes;
    }

    /**
     * Get delivery confirmation level based on origin/destination
     * Return null if delivery confirmation is not acceptable
     *
     * @return int|null
     * @var string $countyDest
     */
    protected function _getDeliveryConfirmationLevel($countyDest = null)
    {
        if (is_null($countyDest)) {
            return null;
        }

        if ($countyDest == Mage_Usa_Model_Shipping_Carrier_Abstract::USA_COUNTRY_ID) {
            return self::DELIVERY_CONFIRMATION_PACKAGE;
        }

        return self::DELIVERY_CONFIRMATION_SHIPMENT;
    }

    /**
     * Get Container Types, that could be customized for UPS carrier
     *
     * @return array
     */
    public function getCustomizableContainerTypes()
    {
        $result = array();
        $containerTypes = $this->getCode('container');
        foreach (parent::getCustomizableContainerTypes() as $containerType) {
            $result[$containerType] = $containerTypes[$containerType];
        }
        return $result;
    }

    /**
     * Return structured data of containers witch related with shipping methods
     *
     * @return array|bool
     */
    public function getContainerTypesFilter()
    {
        return $this->getCode('containers_filter');
    }

    /**
     * Get tracking
     *
     * @param mixed $trackings
     * @return mixed
     */
    public function getTracking($trackings)
    {
        $result = Mage::getModel('shipping/tracking_result');
        $errorTitle = 'Unable to retrieve tracking';
        $resultArr = [];
        $packageProgress = [];

        if (!is_array($trackings)) {
            $trackings = [$trackings];
        }

        foreach ($trackings as $trackingNumber) {
            $debugData = ['request' => null, 'result' => null];

            try {
                $frontendName   = Mage::app()->getStore()->getFrontendName();
                $transactionSrc = is_null($frontendName) ? 'Magento' : $frontendName;
                $headers        = [
                    'Authorization: Bearer ' . $this->getToken(),
                    'transId: ' . $trackingNumber,
                    'transactionSrc: testing' . $transactionSrc
                ];

                $curl = $this->getCurl('api/track/v1/details/' . $trackingNumber . '?' . http_build_query(['returnSignature' => true]), [], $headers);

                $debugData['request'] = json_encode(['url' => curl_getinfo($curl, CURLINFO_EFFECTIVE_URL), 'payload' => [], 'headers' => $headers]);
                $response = curl_exec($curl);
                $debugData['result'] = $response;
                $error = curl_error($curl);
                curl_close($curl);

                if ($error) {
                    $debugData['result'] = [
                        'error' => $error
                    ];
                } else {
                    $response = json_decode($response, true);

                    if (isset($response['response']['errors'])) {
                        foreach ($response['response']['errors'] as $error) {
                            $debugData['result'] = [
                                'error' => $error['message'],
                                'code' => $error['code']
                            ];

                            $error = Mage::getModel('shipping/tracking_result_error');
                            $error->setCarrier(self::CODE);
                            $error->setCarrierTitle($this->getConfigData('title'));
                            $error->setTracking($trackingNumber);
                            $error->setErrorMessage($errorTitle);
                            $result->append($error);
                        }
                    } else {
                        $resultArr['service'] = $response['trackResponse']['shipment'][0]['package'][0]['service']['description'];
                        $resultArr['weight'] = $response['trackResponse']['shipment'][0]['package'][0]['weight']['weight'] . ' ' . $response['trackResponse']['shipment'][0]['package'][0]['weight']['unitOfMeasurement'];

                        $activityTags = $response['trackResponse']['shipment'][0]['package'][0]['activity'];
                        if ($activityTags) {
                            $i = 1;

                            foreach ($activityTags as $activityTag) {
                                $addArr = [];
                                if (isset($activityTag['location']['address']['city'])) {
                                    $addArr[] = $activityTag['location']['address']['city'];
                                }
                                if (isset($activityTag['location']['address']['stateProvinceCode'])) {
                                    $addArr[] = $activityTag['location']['address']['stateProvinceCode'];
                                }
                                if (isset($activityTag['location']['address']['countryCode'])) {
                                    $addArr[] = $activityTag['location']['address']['countryCode'];
                                }

                                $dateArr = [];
                                $date = $activityTag['date'];//YYYYMMDD
                                $dateArr[] = substr($date, 0, 4);
                                $dateArr[] = substr($date, 4, 2);
                                $dateArr[] = substr($date, -2, 2);

                                $timeArr = array();
                                $time = (string)$activityTag['time'];//HHMMSS
                                $timeArr[] = substr($time, 0, 2);
                                $timeArr[] = substr($time, 2, 2);
                                $timeArr[] = substr($time, -2, 2);

                                if ($i == 1) {
                                    $resultArr['status'] = trim($activityTag['status']['description']);
                                    $resultArr['deliverydate'] = implode('-', $dateArr);//YYYY-MM-DD
                                    $resultArr['deliverytime'] = implode(':', $timeArr);//HH:MM:SS
                                    if ($addArr) {
                                        $resultArr['deliveryto'] = implode(', ', $addArr);
                                    }
                                } else {
                                    $tempArr = array();
                                    $tempArr['activity'] = trim($activityTag['status']['description']);
                                    $tempArr['deliverydate'] = implode('-', $dateArr);//YYYY-MM-DD
                                    $tempArr['deliverytime'] = implode(':', $timeArr);//HH:MM:SS
                                    if ($addArr) {
                                        $tempArr['deliverylocation'] = implode(', ', $addArr);
                                    }
                                    $packageProgress[] = $tempArr;
                                }
                                $i++;
                            }
                            $resultArr['progressdetail'] = $packageProgress;
                        }

                        $tracking = Mage::getModel('shipping/tracking_result_status');
                        $tracking->setCarrier(self::CODE);
                        $tracking->setCarrierTitle($this->getConfigData('title'));
                        $tracking->setTracking($trackingNumber);
                        $tracking->addData($resultArr);
                        $result->append($tracking);
                    }
                }
            } catch (Exception $e) {
                $debugData['result'] = [
                    'error' => $e->getMessage(),
                    'code' => $e->getCode()
                ];

                $error = Mage::getModel('shipping/tracking_result_error');
                $error->setCarrier(self::CODE);
                $error->setCarrierTitle($this->getConfigData('title'));
                $error->setTracking($trackingNumber);
                $error->setErrorMessage($errorTitle);
                $result->append($error);
            }
            $this->_debug($debugData);
        }

        return $result;
    }
}
